from tinyalign import edit_distance, hamming_distance

import random
import pytest


STRING_PAIRS = [
    ('', ''),
    ('', 'A'),
    ('A', 'A'),
    ('AB', ''),
    ('AB', 'ABC'),
    ('TGAATCCC', 'CCTGAATC'),
    ('ANANAS', 'BANANA'),
    ('SISSI', 'MISSISSIPPI'),
    ('GGAATCCC', 'TGAGGGATAAATATTTAGAATTTAGTAGTAGTGTT'),
    ('TCTGTTCCCTCCCTGTCTCA', 'TTTTAGGAAATACGCC'),
    ('TGAGACACGCAACATGGGAAAGGCAAGGCACACAGGGGATAGG', 'AATTTATTTTATTGTGATTTTTTGGAGGTTTGGAAGCCACTAAGCTATACTGAGACACGCAACAGGGGAAAGGCAAGGCACA'),
    ('TCCATCTCATCCCTGCGTGTCCCATCTGTTCCCTCCCTGTCTCA', 'TTTTAGGAAATACGCCTGGTGGGGTTTGGAGTATAGTGAAAGATAGGTGAGTTGGTCGGGTG'),
    ('A', 'TCTGCTCCTGGCCCATGATCGTATAACTTTCAAATTT'),
    ('GCGCGGACT', 'TAAATCCTGG'),
    ]


def py_edit_distance(s, t):
    """
    Pure-Python edit distance
    """
    m = len(s)
    n = len(t)

    costs = list(range(m + 1))
    for j in range(1, n + 1):
        prev = costs[0]
        costs[0] += 1
        for i in range(1, m + 1):
            c = min(
                prev + int(s[i-1] != t[j-1]),
                costs[i] + 1,
                costs[i-1] + 1,
            )
            prev = costs[i]
            costs[i] = c

    return costs[-1]


def random_string():
    return ''.join(random.choice('AC') for _ in range(random.randint(0, 20)))


RANDOM_STRING_PAIRS = [(random_string(), random_string()) for _ in range(10000)]


def test_edit_distance():
    assert edit_distance('', '') == 0
    assert edit_distance('', 'A') == 1
    assert edit_distance('A', 'B') == 1
    assert edit_distance('A', 'A') == 0
    assert edit_distance('A', 'AB') == 1
    assert edit_distance('BA', 'AB') == 2
    for s, t in STRING_PAIRS + RANDOM_STRING_PAIRS:
        assert edit_distance(s, '') == len(s)
        assert edit_distance('', s) == len(s)
        assert edit_distance(s, t) == edit_distance(t, s)
        assert edit_distance(s, t) == py_edit_distance(s, t)


def assert_banded(s, t, maxdiff):
    banded_dist = edit_distance(s, t, maxdiff=maxdiff)
    true_dist = edit_distance(s, t)
    if true_dist > maxdiff:
        assert banded_dist > maxdiff
    else:
        assert banded_dist == true_dist


def test_edit_distance_banded():
    for maxdiff in range(5):
        assert_banded('ABC', '', maxdiff)
        for s, t in STRING_PAIRS:
            assert_banded(s, '', maxdiff)
            assert_banded('', s, maxdiff)
            assert_banded(s, t, maxdiff)
            assert_banded(t, s, maxdiff)


def test_hamming_distance():
    assert hamming_distance('', '') == 0
    assert hamming_distance('A', 'A') == 0
    assert hamming_distance('HELLO', 'HELLO') == 0
    assert hamming_distance('ABC', 'DEF') == 3
    assert hamming_distance('ABCXDEF', 'ABCYDEF') == 1


def test_hamming_distance_incorrect_length():
    with pytest.raises(IndexError):
        hamming_distance('A', 'BC')



# cython: language_level=3

from cpython.mem cimport PyMem_Malloc, PyMem_Free

import cython


@cython.wraparound(False)
@cython.boundscheck(False)
def edit_distance(s, t, int maxdiff=-1):
    """
    Return the edit distance between the strings s and t.
    The edit distance is the sum of the numbers of insertions, deletions,
    and mismatches that is minimally necessary to transform one string
    into the other.

    If maxdiff is not -1, then a banded alignment is performed. In that case,
    the true edit distance is returned if and only if it is maxdiff or less.
    Otherwise, a value is returned that is guaranteed to be greater than
    maxdiff, but which is not necessarily the true edit distance.
    """
    cdef:
        unsigned int m = len(s)  # index: i
        unsigned int n = len(t)  # index: j
        int e = maxdiff
        unsigned int i, j, start, stop, c, smallest
        unsigned int prev
        bint match
        bytes s_bytes, t_bytes
        char* sv
        char* tv

    # Return early if string lengths are too different
    cdef unsigned int absdiff = m - n if m > n else n - m
    if e != -1 and absdiff > e:
        return absdiff

    s_bytes = s.encode() if isinstance(s, unicode) else s
    t_bytes = t.encode() if isinstance(t, unicode) else t
    sv = s_bytes
    tv = t_bytes

    # Skip identical prefixes
    while m > 0 and n > 0 and sv[0] == tv[0]:
        sv += 1
        tv += 1
        m -= 1
        n -= 1

    # Skip identical suffixes
    while m > 0 and n > 0 and sv[m-1] == tv[n-1]:
        m -= 1
        n -= 1

    cdef unsigned int result
    cdef unsigned int* costs = <unsigned int*>PyMem_Malloc((m + 1) * sizeof(unsigned int))
    if not costs:
        raise MemoryError()

    with nogil:
        for i in range(m + 1):
            costs[i] = i
        if e == -1:
            # Regular (unbanded) global alignment
            prev = 0
            for j in range(1, n + 1):
                prev = costs[0]
                costs[0] += 1
                for i in range(1, m+1):
                    match = sv[i-1] == tv[j-1]
                    c = 1 + min(
                        prev - match,
                        costs[i],
                        costs[i-1],
                    )
                    prev = costs[i]
                    costs[i] = c
            result = costs[m]
        else:
            # Banded alignment
            smallest = 0
            for j in range(1, n + 1):
                stop = min(j + e + 1, m + 1)
                if j <= e:
                    prev = costs[0]
                    costs[0] += 1
                    smallest = costs[0]
                    start = 1
                else:
                    start = j - e
                    prev = costs[start - 1]
                    smallest = maxdiff + 1
                for i in range(start, stop):
                    match = sv[i-1] == tv[j-1]
                    c = 1 + min(
                        prev - match,
                        costs[i],
                        costs[i-1],
                    )
                    prev = costs[i]
                    costs[i] = c
                    smallest = min(smallest, c)
                if smallest > maxdiff:
                    break
            if smallest > maxdiff:
                result = smallest
            else:
                result = costs[m]
    PyMem_Free(costs)
    return result


def hamming_distance(unicode s, unicode t):
    """
    Compute hamming distance between two strings. If they do not have the
    same length, an IndexError is raised.

    Return the number of differences between the strings.
    """
    cdef Py_ssize_t m = len(s)
    cdef Py_ssize_t n = len(t)
    if m != n:
        raise IndexError("sequences must have the same length")
    cdef Py_ssize_t e = 0
    cdef Py_ssize_t i
    for i in range(m):
        if s[i] != t[i]:
            e += 1
    return e



import sys
import os.path
from setuptools import setup, Extension, find_packages
from distutils.command.sdist import sdist as _sdist
from distutils.command.build_ext import build_ext as _build_ext


def no_cythonize(extensions, **_ignore):
    """Change .pyx to .c or .cpp (copied from Cython documentation)"""
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)
            if ext in (".pyx", ".py"):
                if extension.language == "c++":
                    ext = ".cpp"
                else:
                    ext = ".c"
                sfile = path + ext
            sources.append(sfile)
        extension.sources[:] = sources


extensions = [
    Extension("tinyalign._core", sources=["src/tinyalign/_core.pyx"]),
]


class BuildExt(_build_ext):
    def run(self):
        # If we encounter a PKG-INFO file, then this is likely a .tar.gz/.zip
        # file retrieved from PyPI that already includes the pre-cythonized
        # extension modules, and then we do not need to run cythonize().
        if os.path.exists("PKG-INFO"):
            no_cythonize(extensions)
        else:
            # Otherwise, this is a 'developer copy' of the code, and then the
            # only sensible thing is to require Cython to be installed.
            from Cython.Build import cythonize

            self.extensions = cythonize(self.extensions)
        super().run()


class SDist(_sdist):
    def run(self):
        # Make sure the compiled Cython files in the distribution are up-to-date
        from Cython.Build import cythonize

        cythonize(extensions)
        super().run()


with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="tinyalign",
    setup_requires=["setuptools_scm"],  # Support pip versions that don't know about pyproject.toml
    use_scm_version={"write_to": "src/tinyalign/_version.py"},
    author="Marcel Martin",
    author_email="marcel.martin@scilifelab.se",
    url="https://github.com/marcelm/tinyalign/",
    description="Fast banded edit distance",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    packages=find_packages("src"),
    package_dir={"": "src"},
    ext_modules=extensions,
    cmdclass={"build_ext": BuildExt, "sdist": SDist},
    python_requires='>=3.6',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
)
